;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "toc-org" "20260514.1415"
  "Add table of contents to org-mode files (formerly, org-toc)."
  ()
  :url "https://github.com/snosov1/toc-org"
  :commit "781376e9dc9a901116c0c39914aeb4d46e524e0a"
  :revdesc "781376e9dc9a"
  :keywords '("org-mode" "org-toc" "toc-org" "org" "toc" "table" "of" "contents")
  :authors '(("Sergei Nosov" . "sergei.nosov[at]gmail.com"))
  :maintainers '(("Sergei Nosov" . "sergei.nosov[at]gmail.com")))
